# Generated by Django 3.2.9 on 2022-02-11 06:24

from django.db import migrations, models
import polaris.models


class Migration(migrations.Migration):
    dependencies = [
        ("polaris", "0013_auto_20211011_1956"),
    ]

    operations = [
        migrations.CreateModel(
            name="PolarisHeartbeat",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("key", models.CharField(max_length=80, unique=True)),
                ("last_heartbeat", models.DateTimeField(blank=True, null=True)),
            ],
        ),
        migrations.AddField(
            model_name="transaction",
            name="queue",
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="transaction",
            name="queued_at",
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="transaction",
            name="submission_status",
            field=models.CharField(
                choices=[
                    ("not_ready", "not_ready"),
                    ("ready", "ready"),
                    ("processing", "processing"),
                    ("pending", "pending"),
                    ("pending_funding", "pending_funding"),
                    ("pending_trust", "pending_trust"),
                    ("blocked", "blocked"),
                    ("unblocked", "unblocked"),
                    ("completed", "completed"),
                    ("failed", "failed"),
                ],
                default="not_ready",
                max_length=31,
            ),
        ),
    ]
