# Generated by Django 2.2.24 on 2021-08-17 20:37

from django.db import migrations, models


def match_amount_in(apps, _schema_editor):
    Transaction = apps.get_model("polaris", "Transaction")
    Transaction.objects.update(amount_expected=models.F("amount_in"))


def nop(*_):
    pass


class Migration(migrations.Migration):
    dependencies = [
        ("polaris", "0010_auto_20210809_1605"),
    ]

    operations = [
        migrations.AddField(
            model_name="transaction",
            name="amount_expected",
            field=models.DecimalField(
                blank=True, decimal_places=7, max_digits=30, null=True
            ),
        ),
        migrations.RunPython(match_amount_in, nop),
    ]
