{% if widget.attrs.symbol %}
<div class="icon icon-symbol">{{ widget.attrs.symbol }}</div>
{% endif %}
<input class="form-control {{widget.attrs.class}}" type="{{ widget.type }}" name="{{ widget.name }}" placeholder="{% if 'polaris-transaction-form-amount' in widget.attrs.class and widget.attrs.placeholder == None %}0{% else %}{{widget.attrs.placeholder|stringformat:'s'|default:' '}}{% endif %}" {% if widget.value != None %} value="{{ widget.value|stringformat:'s' }}" {% endif %} {% include "django/forms/widgets/attrs.html" %} >
