# Generated by Django 2.2.18 on 2021-03-01 18:28

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("polaris", "0006_transaction_more_info_url"),
    ]

    operations = [
        migrations.AddField(
            model_name="transaction",
            name="on_change_callback",
            field=models.TextField(blank=True, null=True),
        ),
    ]
